package jatva.lang.synchronization;

import jatva.lang.InvariantException;
import jatva.lang.ModelEvent;
import jatva.lang.MultipleCorrectStatesException;
import jatva.lang.NoCorrectSeriesException;
import jatva.lang.StateChangedDuringFrameException;
import jatva.tracer.messages.ModelMessage;
import jatva.util.Utils;

import java.util.List;

import com.unitesk.jogs.Snapshot;
import com.unitesk.jogs.SnapshotServices;

/**
 * @author <a href="mailto:monakhov@ispras.ru">Alexander Monakhov</a>
 */
public abstract class AbstractSerializer
{
    public static class SerializerConfig
    {
        public boolean seekAllSeries = false;
        public boolean checkStateNotChangedInFrame = false;
    }

    protected SerializerConfig config;

    public void setConfig( SerializerConfig config )
    { this.config = config; }

    protected EventPool eventPool = null;

    protected SnapshotServices snapshotServices = null;

    /**
     * The state of the model after the first successful serie.
     */
    protected Snapshot endModelState = null;

    protected TestManager registrar;

    public void traceModelMessage( String info )
    {
        ModelMessage mm = new ModelMessage( registrar, info );
        mm.makeStringForm();
        mm.trace();
    }

    protected abstract SerializationResult serialize( Snapshot testStateBeforeFrameSnapshot );

    public void performAnalysis( EventPool ep
                               , Snapshot testStateBeforeFrame
                               , SnapshotServices snapshotServices
                               )
    {
        this.snapshotServices = snapshotServices;

        if( config.checkStateNotChangedInFrame )
        {
            if( !(snapshotServices.isRootEquiv( testStateBeforeFrame )) )
                throw new StateChangedDuringFrameException();
        }

        this.eventPool = ep;
        this.registrar = TestManager.getRegistrar();

        traceModelMessage( "serializer: events: " + Utils.endl + ep.toString() );

        switch( serialize( testStateBeforeFrame ) )
        {
            case NO_CORRECT_SERIES:
                traceModelMessage("serializer: no correct series found");
                throw new NoCorrectSeriesException();

            case END_STATES_ARE_NOT_EQUIVALENT:
                traceModelMessage("serializer: found several correct series with different end states");
                throw new MultipleCorrectStatesException();
        }
    }

    public static enum SerializationResult
    {
          OK
        , END_STATES_ARE_NOT_EQUIVALENT
        , NO_CORRECT_SERIES
    }

    protected boolean checkModelEvent( ModelEvent me )
    {
        try
        { return me.check(); }
        catch( InvariantException ie )
        { return false; }
    }

    /**
     * Is the event being currently checked the last event in the sequence.
     *
     */
    protected boolean isCurrentEventTheLastInSequence = false;

    public boolean eventIsTheLastOneInSequence()
    { return isCurrentEventTheLastInSequence; }

    protected String printSerie( String prefix, List<ModelEvent> serie )
    {
        StringBuffer s = new StringBuffer(30);

        s.append(prefix);
        s.append( Utils.endl );
        if( serie.size() == 0 )
        {
            s.append( "{}" );
            return s.toString();
        }
        s.append( "{" );
        for(ModelEvent me : serie )
        {
          s.append( Utils.endl );
          s.append("[");
          s.append( me.getChannelId() );
          s.append(", ");
          s.append( me.getNumberInChannel() );
          s.append( "] :" );
          s.append( Utils.endl );
          s.append( "    " );
          s.append( me.toString() );
        }
        s.append( Utils.endl );
        s.append( "}" );
        return s.toString();
    }

    public abstract AbstractSerializer clone();
}
