/*
 * Copyright (c) 2004 ISP RAS. 109004, B. Communisticheskaya, 25, Moscow,
 * Russia. All rights reserved.
 *
 * $Id: EventPool.java,v 1.9 2009/03/28 13:03:31 almo Exp $ Created on
 * 11.02.2005
 */
package jatva.lang.synchronization;

import jatva.lang.ModelEvent;
import jatva.lang.patterns.ModelEventPattern;
import jatva.util.Utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;


/**
 * @author Victor Kuliamin
 * @author <a href="mailto:monakhov@ispras.ru">Alexander Monakhov</a>
 */
public class EventPool
{
    Hashtable<Object, LinkedList<ModelEvent>> items;

    ArrayList<Object> channelsList;

    public EventPool()
    {
        this( 1 );
    }

    public EventPool( int channels )
    {
        items = new Hashtable<Object, LinkedList<ModelEvent>>( channels );
        channelsList = new ArrayList<Object>( channels );
    }

    public void clear()
    {
        items.clear();
        channelsList.clear();
        listenersQueuesMap.clear();
        size = 0;
    }

    public int size()
    {
        return size;
    }

    public int channels()
    {
        return channelsList.size();
    }

    ConcurrentHashMap<Thread, BlockingQueue<ModelEvent>> listenersQueuesMap = new ConcurrentHashMap<Thread, BlockingQueue<ModelEvent>>();

    public synchronized void addEvent( ModelEvent e )
    {
        Object channelId = e.getChannelId();
//      SimpleConsoleLog.getLog().print( "registering", e.toString() );

        // adding an event
        LinkedList<ModelEvent> channel = null;
        if( !( items.containsKey( channelId ) ) )
        {
            channelsList.add( channelId );
            channel = new LinkedList<ModelEvent>();
            items.put( channelId, channel );
        }
        else
            channel = items.get( channelId );

        e.setNumberInChannel( channel.size() );

        channel.addLast( e );

        size++;

        if( listenersQueuesMap.size() == 0 )
            return;

        for( BlockingQueue<ModelEvent> listenerQueue : listenersQueuesMap.values() )
        {
            try
            {
                listenerQueue.put( e );
//              SimpleConsoleLog.getLog().print( "queue size: " + listenerQueue.size() );
            }
            catch( InterruptedException ie )
            {
//              SimpleConsoleLog.getLog().print( "event enqueue interrupted" );
            }
        }
    }


    /*
     * public void addEvent(ModelEvent e) {
     * //if(TestManager.getDefaultEventDistribution()) { EventList[] newItems =
     * new EventList[channels+1]; System.arraycopy(items, 0, newItems, 0,
     * channels); newItems[channels] = new EventList(); channels++; items =
     * newItems; items[channels-1].addLast(e); } //else { //items[0].addLast(e);
     * } size++; }
     */

    /*
     * public void removeEvents(int ch, int num) { if(ch < channels && ch >= 0)
     * { EventList l = items[ch]; if(l.size() > num) { for(int i = 0; i < num;
     * i++) l.removeLast(); } if(l.size() <= num) l.clear(); } }
     */

    public LinkedList<ModelEvent> getChannel( int i )
    {
        return items.get( channelsList.get( i ) );
    }

    @Override
    public String toString()
    {
        StringBuffer r = new StringBuffer( "{" );
        for( Object channelId : items.keySet() )
        {
            r.append( Utils.endl );
            r.append( channelId );
            r.append( ":" );
            List<ModelEvent> events_in_channel = items.get( channelId );
            for( ModelEvent me : events_in_channel )
            {
                r.append( Utils.endl );
                r.append( "    " );
                r.append( me.toString( false ) );
            }
        }
        r.append( Utils.endl );
        r.append( "}" );
        return r.toString();
    }

    /*
     * boolean containsKey(Object key); boolean containsValue(Object value);
     * boolean contains(Object key, Object val);
     *
     * Object get(Object key);
     *
     * Object removeKey(Object key); boolean removeValue(Object val); boolean
     * remove(Object key, Object val);
     *
     * Object put(Object key, Object value);
     *
     * Object clone(); String toString(); boolean equals(Object o); boolean
     * equalsByRefs(Object o);
     */

    private int size = 0;

    // private EventList[] items;

    // public void unboundAllEvents()
    // {
    // for( LinkedList<ModelEvent> events :items.values() )
    // {
    // for( ModelEvent me : events )
    // me.unbound();
    // }
    // }

    /**
     * Initializes event waiting functionality for the current thread.
     *
     * <p>If event waiting was not initialized by this thread earlier, then
     * this method creates a new queue in {@link #listenersQueuesMap}. Otherwise,
     * the existing queue is cleared.</p>
     */
    public void initEventWaiting()
    {
        Thread curr = Thread.currentThread();

        synchronized( listenersQueuesMap )
        {
            if( listenersQueuesMap.containsKey( curr ) )
            {
                listenersQueuesMap.get( curr ).clear();
            } else {
                listenersQueuesMap.put( curr, new LinkedBlockingQueue<ModelEvent>() );
            }
        }
    }

    public void stopEventWaiting()
    {
        synchronized( listenersQueuesMap )
        {
            listenersQueuesMap.remove( Thread.currentThread() );
        }
    }

    public EventWaitingResult waitForEvent( long millis, ModelEventPattern[] patterns )
    {
        long start_time = System.currentTimeMillis();

        Thread curr = Thread.currentThread();

        BlockingQueue<ModelEvent> queue = null;

        synchronized( listenersQueuesMap )
        {
            if( listenersQueuesMap.containsKey( curr ) )
                queue = listenersQueuesMap.get( curr );
            else
                listenersQueuesMap.put( curr, queue = new LinkedBlockingQueue<ModelEvent>() );
        }

//        int result = Integer.MAX_VALUE;

//        EVENT_LOOP:
        while( true )
        {
            long curr_time = System.currentTimeMillis();
            long time_left = millis - ( curr_time - start_time );
            if( time_left < 0 )
            {
//                result = -1;
//                break;
                return EventWaitingResult.TIMEOUT_EXPIRED;
            }

            ModelEvent me = null;
            try
            {
                me = queue.poll( time_left, TimeUnit.MILLISECONDS );
            }
            catch( InterruptedException e )
            {}

            if( me == null )
            {
                continue;
            }

            for( int i = 0; i < patterns.length; i ++ )
            {
                if( patterns[ i ].holdsOn( me ) )
                {
                    return new EventWaitingResult( i, me );
//                    result = i;
//                    break EVENT_LOOP;
                }
            }
        }

//        return result;
    }

    public int getSize()
    {
        return size;
    }

    public Collection<ModelEvent> next( ModelEvent me )
    {
        LinkedList<ModelEvent> channel = items.get( me.getChannelId() );
        if( me.getNumberInChannel() == channel.size() - 1 )
            return null;
        else
            return Collections.singleton( channel.get( me.getNumberInChannel() + 1 ) );
    }

    public Hashtable<Object, LinkedList<ModelEvent>> getItems()
    {
        return items;
    }
}
