package jatva.lang.synchronization;

import jatva.lang.ModelObject;
import jatva.lang.annotation.JTSInternal;
import jatva.lang.annotation.MediatorClass;
import jatva.lang.annotation.ScenarioClass;

import java.lang.reflect.Field;

import com.unitesk.utils.predicates.Predicate;

/**
 * @author <a href="mailto:monakhov@ispras.ru">Alexander Monakhov</a>
 */
public class JTSFieldSkipFilter implements Predicate<Field>
{
    public static final JTSFieldSkipFilter instance  = new JTSFieldSkipFilter();

    private JTSFieldSkipFilter(){}

    @Override
    public boolean holdsOn( Field f )
    {
        // filter out all jts internal fields
        if( f.getAnnotation( JTSInternal.class ) != null )
            return true;

        Class<?> decl_class = f.getDeclaringClass();

        if( decl_class == Throwable.class )
        {
            String field_name = f.getName();
            return    field_name.equals( "stackTrace" )
                   || field_name.equals( "backtrace" )
                 ;
        }

        // filter out all fields declared in mediator classes
        if( decl_class.getAnnotation( MediatorClass.class ) != null )
            return true;

        // filter out all fields of scenario class that are not
        // references to objects of specification classes
        if( decl_class.getAnnotation( ScenarioClass.class ) != null )
        {
            return !(ModelObject.class.isAssignableFrom( f.getType() ));
        }

        return false;
    }
}
