/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: Serializer.java,v 1.16 2009/04/06 11:10:41 almo Exp $
 * Created on 15.02.2005
 *
 */
package jatva.lang.synchronization;

import jatva.lang.ModelEvent;
import jatva.util.IdentityMap;
import jatva.util.IntList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.unitesk.jogs.Snapshot;

/**
 * @author Victor Kuliamin
 * @author <a href="mailto:monakhov@ispras.ru">Alexander Monakhov</a>
 */
public class Serializer extends AbstractSerializer
{
  //private int       timeDelta = 100;

  private ModelEvent[] sequence = null;
  // the index of the leftmost null element in sequence array.
  private int seq_length = 0;

  private IntList   stack    = new IntList();
  private int[]     used     = null;

  /**
   * The state of the model before the first event registered in the frame.
   */
  private Snapshot initialModelState = null;

//  private IdentityMap storedState = null;
//  private IdentityMap endState    = null;

  private Serializer(){}

  public static final Serializer prototype = new Serializer();

  @Override
  public AbstractSerializer clone()
  { return new Serializer(); }

  protected SerializationResult serialize( Snapshot testStateBeforeFrame )
    {
        sequence = new ModelEvent[ eventPool.size() ];
        Arrays.fill( sequence, null );
        seq_length = 0;
        used = new int[ eventPool.channels() ];

        int num = 0;

        // man.traceModelMessage("serializer: previous event at start: " +
        // previous);
        // man.traceMessage("serializer: number of channels: " +
        // events.channels());
        // man.traceMessage("serializer: number of events: " + events.size());

        this.initialModelState = snapshotServices.makeNewRootSnapshot( testStateBeforeFrame );

        if( !config.seekAllSeries )
        {
            if( makeNextSequence() )
            {
                traceModelMessage( printSerie( "serializer: found series:", serieAsList() ) );

                stack.clear();
                previous = null;
                return SerializationResult.OK;
            }
            else
            {
                stack.clear();
                previous = null;
                return SerializationResult.NO_CORRECT_SERIES;
            }
        }
        else
        {
            if( makeNextSequence() )
            {
                traceModelMessage( printSerie( "serializer: found " + ( ++num ) + " series:", serieAsList() ) );

                endModelState = snapshotServices.makeNewRootSnapshot( testStateBeforeFrame );

                previous = null;
                snapshotServices.restoreRoot( initialModelState );

                while( makeNextSequence() )
                {
                    traceModelMessage( printSerie( "serializer: found " + ( ++num ) + " series:", serieAsList() ) );
                    if( !snapshotServices.isRootEquiv( endModelState ) )
                    {
                        return SerializationResult.END_STATES_ARE_NOT_EQUIVALENT;
                    }

                    snapshotServices.restoreRoot( initialModelState );

                    previous = null;
                }

                snapshotServices.restoreRoot( endModelState );
                traceModelMessage( "serializer: no more series" );

                previous = null;
                return SerializationResult.OK;
            }
            else
            {
                previous = null;
                return SerializationResult.NO_CORRECT_SERIES;
            }
        }
    }

//  public void storeState(IdentityMap s)
//  {
////      System.out.println( "storeState" );
////      System.out.println( "\ncurrent state : " + DeepOperationManager.getDeepString( s ) );
//    storedState = new IdentityMap();
//
//    SynchronizationManager.cloneStarting();
//
//    for(Iterator i = s.keys().iterator(); i.hasNext(); )
//    {
//      DeepOperationManager.getDeepClone(i.next(), storedState);
//    }
//
//    SynchronizationManager.cloneFinished();
//
////    System.out.println( "\nstored state : " + DeepOperationManager.getDeepString( storedState ) );
//    //TestManager.getRegistrar().traceTestMessage("stored state : " + DeepOperationManager.getDeepString(storedState));
//  }

//  public void storeState( IdentityMap s )
//  {
//    storedState = new IdentityMap();
//
//    SynchronizationManager.cloneStarting();
//
//    for(Iterator i = s.keys().iterator(); i.hasNext(); )
//    {
//      DeepOperationManager.getDeepClone(i.next(), storedState);
//    }
//
//    SynchronizationManager.cloneFinished();
//
//  }

//  public void restoreState(IdentityMap s)
//  {
//      System.out.println( "restoreState");
//      System.out.println( "\ncurrent state : " + DeepOperationManager.getDeepString(s) );
//      System.out.println( "\nstored state : " + DeepOperationManager.getDeepString( storedState ) );

    //TestManager.getRegistrar().traceTestMessage("current state : " + DeepOperationManager.getDeepString(s));

//      DeepOperationManager.restoreStarting();
//    for(Iterator i = s.keys().iterator(); i.hasNext(); )
//    {
//      DeepOperationManager.restoreFromClone(i.next(), storedState);
//    }
//    DeepOperationManager.restoreFinished();

//      snapshotServices.restoreRoot( storedState );
//    System.out.println( "\nrestored state : " + DeepOperationManager.getDeepString(s) );
    //TestManager.getRegistrar().traceTestMessage("restored state : " + DeepOperationManager.getDeepString(s));
//  }

//  public void storeEndState(IdentityMap s)
//  {
//    endState = new IdentityMap();
//    for(Iterator i = s.keys().iterator(); i.hasNext(); )
//    {
//      DeepOperationManager.getDeepClone(i.next(), endState);
//    }
//
//    //TestManager.getRegistrar().traceTestMessage("end state : " + DeepOperationManager.getDeepString(endState));
//  }

//  public void restoreEndState(IdentityMap s)
//  {
//    //TestManager.getRegistrar().traceTestMessage("current state : " + DeepOperationManager.getDeepString(s));
//
//      DeepOperationManager.restoreStarting();
//    for(Iterator i = s.keys().iterator(); i.hasNext(); )
//    {
//      DeepOperationManager.restoreFromClone(i.next(), endState);
//    }
//    DeepOperationManager.restoreFinished();
//
//    //TestManager.getRegistrar().traceTestMessage("restored state : " + DeepOperationManager.getDeepString(s));
//  }

//  public boolean compareStates(IdentityMap s)
//  {
//    return DeepOperationManager.areStatesDeeplyEqual(s, endState);
//  }

  public boolean constructSequence(IdentityMap st)
  {
    boolean res = false;
    IntList stack = new IntList();

    // can process < 1024 events in one channel

    int position = 0;
    int t        = 0;
    int i = 0, num = 0;
    boolean tmp = false;

    for(i = 0; i < eventPool.channels(); i++) used[i] = -1;

    if(!config.seekAllSeries)
    {
      while(position < eventPool.size())
      {
        ModelEvent e = eventPool.getChannel(channel(t)).get(index(t));
        sequence[ seq_length ++ ] = e;
        tmp = (checkSequence(/*st*/) == -1);

        if(tmp)
        {
          stack.addLast(t);
          position++;
          used[channel(t)]++;

          i = getNext(-1, eventPool.channels(), eventPool, used);

          if(i < 0 && position == eventPool.size()) break;

          s2:while(i < 0 && position > 0)
          {
            sequence[ seq_length -- ] = null;
            stack.removeLast();
            position--;
            used[channel(t)]--;

            i = getNext(t, eventPool.channels(), eventPool, used);

            if(i >= 0) break s2;
            t = stack.last();
          }

          if(position <= 0) return false;
          else              t = i;
        }
        else
        {
          sequence[ seq_length -- ] = null;
          i = getNext(t, eventPool.channels(), eventPool, used);

          while(i < 0 && position > 0)
          {
            sequence[ seq_length -- ] = null;
            stack.removeLast();
            position--;
            used[channel(t)]--;
            t = stack.last();

            i = getNext(t, eventPool.channels(), eventPool, used);
          }

          if(position <= 0) return false;
          else              t = i;
        }
      }
    }
    else // seekAll == true
    {
      while(true)
      {
        while(position < eventPool.size())
        {
          ModelEvent e = eventPool.getChannel(channel(t)).get(index(t));
          sequence[ seq_length ++ ] = e;
          tmp = (checkSequence(/*st*/) == -1);

          if(tmp)
          {
            stack.addLast(t);
            position++;
            used[channel(t)]++;
            i = getNext(-1, eventPool.channels(), eventPool, used);

            if(i < 0/*i == channels*/ && position == eventPool.size())
            {
              break;
            }

            s2:while(i < 0/*i >= channels*/ && position > 0)
            {
              sequence[ seq_length -- ] = null;
              stack.removeLast();
              position--;
              used[channel(t)]--;

              i = getNext(t, eventPool.channels(), eventPool, used);
              if(i >= 0) break s2;
              t = stack.last();
            }

            if(position <= 0 && i < 0/*i >= channels*/)
            {
              return false;
            }
            else
            {
              t = i;//(i << 10) + j;
            }
          }
          else
          {
            sequence[ seq_length -- ] = null;
            i = getNext(t, eventPool.channels(), eventPool, used);

            while(i < 0/*i >= channels*/ && position > 0)
            {
              t = stack.last();
              sequence[ seq_length -- ] = null;
              stack.removeLast();
              position--;
              used[channel(t)]--;

              i = getNext(t, eventPool.channels(), eventPool, used);
            }

            if(position <= 0 && i < 0/*i >= channels*/)
            {
              return false;
            }
            else
            {
              t = i;//(i << 10) + j;
            }
          }
        }

        System.out.println("" + (++num) + " sequence: " + sequence);
        i = -1;//channels;

        while(i < 0/*i >= channels*/ && position > 0)
        {
          t = stack.last();
          sequence[ seq_length -- ] = null;
          stack.removeLast();
          position--;
          used[channel(t)]--;
          i = getNext(t, eventPool.channels(), eventPool, used);
        }

        if(position <= 0 && i < 0/*i >= channels*/)
        {
          return false;
        }
        else
        {
          t = i;//(i << 10) + j;
        }
      }
    }

    return res;
  }



  public boolean makeNextSequence(/*IdentityMap st*/)
  {
    // can process <= 65536 channels and events in one channel
    int i = 0, j = 0, e = -1;


    if(stack.isEmpty())
    {
      //TestManager man = TestManager.getRegistrar();
      //man.traceMessage("makeNextSequence: found " + events.channels() + " channels");
      for(i = 0; i < eventPool.channels(); i++)
      {
        used[i] = eventPool.getChannel(i).size()-1;
        //man.traceMessage("makeNextSequence: in " + i + "-th channel found " + events.getChannel(i).size() + " events");
        for(j = 0; j < eventPool.getChannel(i).size(); j++)
          stack.addLast((i << 16) + j);
      }

      if((e = checkSequence(/*st*/)) == -1)
      {
        return true;
      }


      e++;
    }
    else e = eventPool.size();

    while(true)
    {
      do
      {
        e--;
        for(i = stack.size()-1; i >= e; i--)
        {
          j = stack.last() >> 16;
          used[j]--;
          stack.removeLast();
        }

        for(i = j + 1; i < eventPool.channels(); i++)
          if(used[i] + 1 < eventPool.getChannel(i).size()) break;
      }
      while(e != -1 && i == eventPool.channels());

      if(e == -1)
      {
        return false;
      }

      while(true)
      {

        stack.addLast((i << 16) + (++used[i]));

        e++;
        if(e >= eventPool.size())
        {
          if((e = checkSequence(/*st*/)) == -1)
          {

            return true;
          }
          else
          {
            e++;
            break;
          }
        }
        else
        {
          for(i = 0; i < eventPool.channels(); i++)
            if(used[i] + 1 < eventPool.getChannel(i).size()) break;

          if(i == eventPool.channels()) { e--; break; }
        }
      }
    }
  }

  public final static int channel(int t) { return t >> 16; }
  public final static int index  (int t) { return t & 65535; }
  public final static int composite(int t, int m) { return (t << 16) + m; }

  private final int getNext(int t, int channels, EventPool events, int[] used)
  {
    for(int i = (t < 0)?0:(channel(t)+1); i < channels; i++)
      if(used[i] + 1 < events.getChannel(i).size()) return (i << 10) + used[i] + 1;

    return -1;
  }

    public int checkSequence(/* IdentityMap s */)
    {
        int t;
        ModelEvent e;
        TestManager mgr = registrar;

        mgr.serializationStart();

        // sequence.clear();
        /*
         * StringBuffer str = new StringBuffer("checking sequence: "); for(int i
         * = 0; i < stack.size(); i++) { t = stack.get(i); str.append(" [" +
         * channel(t) + ", " + index(t) + "] : <" +
         * events.getChannel(channel(t)).get(index(t)) + ">"); }
         */
        // str.append(System.getProperty("line.separator"));
        // mgr.traceModelMessage(printStackWithPrefix("serializer: checking sequence:"));
        ArrayList<ModelEvent> seq = new ArrayList<ModelEvent>();

        try
        {
            for( int i = 0; i < stack.size(); i++ )
            {
                t = stack.get( i );
                e = eventPool.getChannel( channel( t ) ).get( index( t ) );
                // sequence.addLast(e);

                if( stack.size() == eventPool.size() && i == ( stack.size() - 1 ) )
                {
                    isCurrentEventTheLastInSequence = true;
                }
                mgr.setCurrentEvent( e );

                seq.add( e );

                // mgr.traceModelMessage("previous: " + e);

                if( previous != null && !checkImmediateResult( e ) )
                {
                    previous = null;

                    snapshotServices.restoreRoot( initialModelState );
                    mgr.removeCurrentEvent();

                    mgr.serializationEnd();
                    return i;
                }
                else if( previous != null )
                {
                    previous = null;
                }
                else if( !checkModelEvent( e ) )
                {

                    snapshotServices.restoreRoot( initialModelState );
                    mgr.removeCurrentEvent();

                    mgr.serializationEnd();
                    return i;
                }

                // mgr.traceModelMessage("previous: " + e);

                mgr.removeCurrentEvent();
            }

            if( previous != null )
            {
                previous = null;

                mgr.serializationEnd();
                return stack.size() - 1;
            }

            mgr.serializationEnd();
            return -1;
        }
        finally
        {
            isCurrentEventTheLastInSequence = false;
            // String res = "";
            // for( ModelEvent me : seq )
            // {
            // res += me.getOperation().getMethod().getName() + ";";
            // }
            // System.out.println( res );
        }

        /*
         * boolean flag31 = false; boolean flag1_1 = false; boolean flag21_01 =
         * false;
         *
         * for(int i = 0; i < sequence.size(); i++) { if(!flag31 &&
         * sequence.get(i).kind().startsWith("1")) return i; else
         * if(sequence.get(i).kind().startsWith("3")) flag31 = true;
         *
         * if(!flag1_1 && sequence.get(i).kind().endsWith("1")) return i; else
         * if(sequence.get(i).kind().startsWith("1")) flag1_1 = true;
         *
         * if(!flag21_01 && sequence.get(i).kind().startsWith("01")) return i;
         * else if(sequence.get(i).kind().startsWith("21")) flag21_01 = true; }
         *
         * if(sequence.size() > 2 && !sequence.get(2).kind().startsWith("21"))
         * return 2;
         *
         * if(sequence.size() > 5 && !sequence.get(5).kind().startsWith("10"))
         * return 5;
         */
    }

  private ModelEvent previous = null;

  public void markImmediateResult(ModelEvent e)
  {
    previous = e;
  }

  private boolean checkImmediateResult(ModelEvent e)
  {
    if(e.getManager() != previous.getManager())
    {
      TestManager.getRegistrar().traceModelMessage("NO!!! Next event is not the resulting event of the previous call");
      return false;
    }

    TestManager.getRegistrar().traceModelMessage("YES!!! Next event is the resulting event of the previous call");
    return true;
  }

//  private String printStackWithPrefix(String prefix)
//  {
//    StringBuffer s = new StringBuffer(30);
//
//    s.append(prefix);
//    s.append( Utils.endl );
//    s.append( "{" );
//    int t;
//    for(int i = 0; i < stack.size(); i++)
//    {
//      t = stack.get(i);
//      s.append( Utils.endl );
//      s.append("[");
//      s.append(channel(t));
//      s.append(", ");
//      s.append(index(t));
//      s.append( "] :" );
//      s.append( Utils.endl );
//      s.append( "    " );
//      s.append( eventPool.getChannel(channel(t)).get(index(t)).toString( true ) );
//    }
//    s.append( Utils.endl );
//    s.append( "}" );
//    return s.toString();
//  }

  private List<ModelEvent> serieAsList()
  {
      for( int i = 0; i < stack.size(); i ++ )
      {
          int t = stack.get( i );
          sequence[ i ] = eventPool.getChannel( channel(t) ).get( index(t) );
      }

      return Arrays.asList( sequence );
  }

    /*
  public static void xmain(String[] args)
  {
    Serializer s = new Serializer();
    EventPool     m = new EventPool(4);

    m.addEvent(new ModelEvent("00"), 0);
    m.addEvent(new ModelEvent("01"), 0);
    m.addEvent(new ModelEvent("02"), 0);
    m.addEvent(new ModelEvent("10"), 1);
    m.addEvent(new ModelEvent("11"), 1);
    m.addEvent(new ModelEvent("20"), 2);
    m.addEvent(new ModelEvent("21"), 2);
    m.addEvent(new ModelEvent("30"), 3);

    System.out.println(m.toString());

    s.init(m);

    s.constructSequence();

  }*/

  /*
  public static void ymain(String[] args)
  {
    Serializer s = new Serializer();
    EventPool     m = new EventPool(4);

    m.addEvent(new ModelEvent("00"), 0);
    m.addEvent(new ModelEvent("01"), 0);
    m.addEvent(new ModelEvent("02"), 0);
    m.addEvent(new ModelEvent("10"), 1);
    m.addEvent(new ModelEvent("11"), 1);
    m.addEvent(new ModelEvent("20"), 2);
    m.addEvent(new ModelEvent("21"), 2);
    m.addEvent(new ModelEvent("30"), 3);

    System.out.println(m.toString());

    s.init(m);

    int num = 0;
    while(s.makeNextSequence())
    {
      System.out.println("" + (++num) + " sequence: " + s.sequence);
    }

    System.out.println("no more sequences");
  }
  */
}