package jatva.util.log;

/**
 * Synchronized log able to print an array of string atomically.
 *
 * Each block of strings is prefaced with a header and finished with a footer.
 * Ancestor classes are meant to use some identification of the thread executing
 * the printing in the header (or footer).
 * @author <a href="mailto:monakhov@ispras.ru">Alexander Monakhov</a>
 */
public abstract class ThreadLog
{
    protected abstract void printHeader();

    protected abstract void printLine( String l );

    protected abstract void printFooter();

    /**
     * Prints all the strings individually, overall string
     * output is prefaced with {@link #printHeader() header} and
     * closed with {@link #printFooter() footer}.
     *
     * Each string is splitted further with new line character '\n'.
     */
    public synchronized void print( String ... texts )
    {
        printHeader();
        for( String t : texts )
        {
            String[] lines = t.split( "\n" );
            for( String l : lines )
            {
                printLine( l );
            }
        }
        printFooter();
    }
}
